"""All constants for Energy Assistant."""

from typing import Final

ROOT_LOGGER_NAME: Final[str] = "energy_assistant"

DEFAULT_NOMINAL_POWER = 100
DEFAULT_NOMINAL_DURATION = 3600

POWER_HYSTERESIS = 0.1
