"""Extend Devices Fields

Revision ID: 2558c9508fcc
Revises: 2dbe1d6a4fa6
Create Date: 2023-12-16 19:18:13.051874

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "2558c9508fcc"
down_revision = "2dbe1d6a4fa6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("devices", sa.Column("type", sa.String(), nullable=True))
    op.add_column("devices", sa.Column("config", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("devices", "config")
    op.drop_column("devices", "type")
    # ### end Alembic commands ###
